@@ -1,6 +1,6 @@
 # CSV -- module for generating/parsing CSV data.
   
-# $Id: csv.rb,v 1.9 2002/02/27 16:33:46 nahi Exp $
+# $Id: csv.rb,v 1.10 2002/04/13 10:33:53 nahi Exp $
   
 # This module is copyrighted free software by NAKAMURA, Hiroshi.
 # You can redistribute it and/or modify it under the same term as Ruby.
@@ -131,7 +131,6 @@ public
       end
       return true
     end
-  
   end
 
 
@@ -172,8 +171,7 @@ public
   #   if row2.empty?
   #     p "row2 not find."
   #   end
-  #   # You cannot close "csvfile.csv" with this pattern.  Use EXAMPLE 2 or
-  #   # CSV::Reader directly.
+  #   reader.close
   #   
   # EXAMPLE 2
   #   CSV.open( "csvfile.csv", "r" ) do | row |
@@ -183,8 +181,7 @@ public
   # EXAMPLE 3
   #   writer = CSV.open( "csvfile.csv", "w" )
   #   writer << [ "r1c1", "r1c2" ] << [ "r2c1", "r2c2" ] << [ nil, nil ]
-  #   # You cannot close "csvfile.csv" with this pattern.  Use EXAMPLE 4 or
-  #   # CSV::Writer directly.
+  #   writer.close
   #   
   # EXAMPLE 4
   #   CSV.open( "csvfile.csv", "w" ) do | writer |
@@ -193,11 +190,11 @@ public
   #     writer << [ nil, nil ]
   #   end
   #
-  def CSV.open( filename, mode, &block )
+  def CSV.open( filename, mode, colSep = ?,, &block )
     if mode == "r" or mode == "rb"
-      openReader( filename, &block )
+      openReader( filename, colSep, &block )
     elsif mode == "w" or mode == "wb"
-      openWriter( filename, &block )
+      openWriter( filename, colSep, &block )
     else
       raise ArgumentError.new( "'mode' mustbe 'r' or 'w'" )
     end
@@ -206,28 +203,39 @@ public
   # Private class methods.
   class << self
   private
-    def openReader( filename, &block )
+    def openReader( filename, colSep, &block )
       file = File.open( filename, "rb" )
       if block
-	CSV::Reader.parse( file ) do | row |
-	  yield( row )
+	begin
+	  CSV::Reader.parse( file, colSep ) do | row |
+	    yield( row )
+	  end
+	ensure
+	  file.close
 	end
-	file.close
 	nil
       else
-	CSV::Reader.create( file )
+	reader = CSV::Reader.create( file, colSep )
+	reader.closeOnTerminate
+	reader
       end
     end
 
-    def openWriter( filename, &block )
+    def openWriter( filename, colSep, &block )
       file = File.open( filename, "wb" )
       if block
-	CSV::Writer.generate( file ) do | writer |
-	  yield( writer )
+	begin
+	  CSV::Writer.generate( file, colSep ) do | writer |
+	    yield( writer )
+	  end
+	ensure
+	  file.close
 	end
-	file.close
+	nil
       else
-	CSV::Writer.create( file )
+	writer = CSV::Writer.create( file, colSep )
+	writer.closeOnTerminate
+	writer
       end
     end
   end
@@ -266,11 +274,12 @@ public
     #   Block value is always nil.  Rows are not cached for performance
     #   reason.
     #
-    def Reader.parse( stringOrReadable )
-      reader = create( stringOrReadable )
+    def Reader.parse( stringOrReadable, colSep = ?, )
+      reader = create( stringOrReadable, colSep )
       reader.each do | row |
 	yield( row )
       end
+      reader.close
       nil
     end
 
@@ -286,14 +295,14 @@ public
     # DESCRIPTION
     #   Create instance.  To get parse result, see CSV::Reader#each.
     #   
-    def Reader.create( stringOrReadable )
+    def Reader.create( stringOrReadable, colSep = ?, )
       case stringOrReadable
       when IO
-	IOReader.new( stringOrReadable )
+	IOReader.new( stringOrReadable, colSep )
       when String
-	StringReader.new( stringOrReadable )
+	StringReader.new( stringOrReadable, colSep )
       else
-	IOReader.new( stringOrReadable )
+	IOReader.new( stringOrReadable, colSep )
       end
     end
 
@@ -342,6 +351,19 @@ public
       row
     end
 
+    # SYNOPSIS
+    #   CSV::Reader#close
+    #
+    # RETURNS
+    #   nil
+    #
+    # DESCRIPTION
+    #   Close this reader.
+    #   
+    def close
+      terminate
+    end
+
   private
     def initialize( dev )
       raise RuntimeError.new( 'Do not instanciate this class directly.' )
@@ -350,11 +372,15 @@ public
     def getRow( row )
       raise NotImplementedError.new( 'Method getRow must be defined in a derived class.' )
     end
+
+    def terminate
+      # Define if needed.
+    end
   end
   
 
   # DESCRIPTION
-  #   CSV::IOReader -- CSV formatted stream reader.
+  #   CSV::StringReader -- CSV formatted stream reader.
   #   
   # EXAMPLE
   #   Read CSV lines untill the first column is 'stop'.
@@ -379,14 +405,15 @@ public
     # DESCRIPTION
     #   Create instance.  To get parse result, see CSV::Reader#each.
     #   
-    def initialize( string )
+    def initialize( string, colSep = ?, )
+      @colSep = colSep
       @dev = string
       @idx = 0
     end
 
   private
     def getRow( row )
-      nofCells, nextIdx = CSV.parseLine( @dev, @idx, row )
+      nofCells, nextIdx = CSV.parseLine( @dev, @idx, row, @colSep )
       if nofCells == 0 && nextIdx == 0 && @idx != @dev.size
 	raise IllegalFormatError.new
       end
@@ -422,14 +449,31 @@ public
     # DESCRIPTION
     #   Create instance.  To get parse result, see CSV::Reader#each.
     #   
-    def initialize( io )
-      @dev = CSV::IOBuf.new( io )
+    def initialize( io, colSep = ?, )
+      @io = io
+      @colSep = colSep
+      @dev = CSV::IOBuf.new( @io )
       @idx = 0
+      @closeOnTerminate = false
+    end
+
+    # SYNOPSIS
+    #   CSV::IOReader#closeOnTerminate
+    #
+    # RETURNS
+    #   true
+    #
+    # DESCRIPTION
+    #   Tell this reader to close the IO when terminated (Triggered by invoking
+    #   CSV::IOReader#close).
+    #   
+    def closeOnTerminate
+      @closeOnTerminate = true
     end
 
   private
     def getRow( row )
-      nofCells, nextIdx = CSV.parseLine( @dev, @idx, row )
+      nofCells, nextIdx = CSV.parseLine( @dev, @idx, row, @colSep )
       if nofCells == 0 && nextIdx == 0 && !@dev.isEOS?
 	raise IllegalFormatError.new
       end
@@ -437,6 +481,16 @@ public
       @idx = nextIdx - dropped
       return nofCells
     end
+
+    def terminate
+      if @closeOnTerminate
+	@io.close
+      end
+
+      if @dev
+	@dev.close
+      end
+    end
   end
 
 
@@ -486,9 +540,10 @@ public
     #   To add CSV data to generate CSV string, see CSV::Writer#<< or
     #   CSV::Writer#addRow.
     #   
-    def Writer.generate( stringOrWritable )
-      writer = Writer.create( stringOrWritable )
+    def Writer.generate( stringOrWritable, colSep = ?, )
+      writer = Writer.create( stringOrWritable, colSep )
       yield( writer )
+      writer.close
       nil
     end
 
@@ -506,8 +561,8 @@ public
     #   Create instance.  To add CSV data to generate CSV string, see
     #   CSV::Writer#<< or CSV::Writer#addRow.
     #   
-    def Writer.create( stringOrReadable )
-      BasicWriter.new( stringOrReadable )
+    def Writer.create( stringOrReadable, colSep = ?, )
+      BasicWriter.new( stringOrReadable, colSep )
     end
 
     # SYNOPSIS
@@ -524,7 +579,7 @@ public
     #   [ "c1", "c2", "c3" ].
     #   
     def <<( row )
-      @dev << CSV.create( row ) << "\r\n"
+      @dev << CSV.create( row, @colSep ) << "\r\n"
       self
     end
 
@@ -543,14 +598,31 @@ public
     #   (Formar is "c1" and latter is Null.)
     #   
     def addRow( row )
-      CSV.createLine( row, row.size, @dev )
+      CSV.createLine( row, row.size, @dev, @colSep )
       self
     end
 
+    # SYNOPSIS
+    #   CSV::Writer#close
+    #
+    # RETURNS
+    #   nil
+    #
+    # DESCRIPTION
+    #   Close this writer.
+    #   
+    def close
+      terminate
+    end
+
   private
     def initialize( dev )
       raise RuntimeError.new( 'Do not instanciate this class directly.' )
     end
+
+    def terminate
+      # Define if needed.
+    end
   end
 
 
@@ -574,16 +646,41 @@ public
     #   Create instance.  To add CSV data to generate CSV string, see
     #   CSV::Writer#<< or CSV::Writer#addRow.
     #   
-    def initialize( stringOrWritable )
+    def initialize( stringOrWritable, colSep = ?, )
+      @colSep = colSep
       @dev = stringOrWritable
+      @closeOnTerminate = false
+    end
+
+    # SYNOPSIS
+    #   CSV::BasicWriter#closeOnTerminate
+    #
+    # RETURNS
+    #   true
+    #
+    # DESCRIPTION
+    #   Tell this writer to close the IO when terminated (Triggered by invoking
+    #   CSV::BasicWriter#close).
+    #   
+    def closeOnTerminate
+      @closeOnTerminate = true
+    end
+
+  private
+    def terminate
+      if @closeOnTerminate
+	@dev.close
+      end
     end
   end
-  
+
   # SYNOPSIS
-  #   cells = CSV.parse( src )
+  #   cells = CSV.parse( src, colSep = ?, )
   #
   # ARGS
   #   src: a CSV String.
+  #   colSep: Column separator.  ?, by default.  If you want to separate
+  #     fields with semicolon, give ?; here.
   #
   # RETURNS
   #   cells: an Array of parsed cells in first line.  Each cell is a String.
@@ -596,16 +693,16 @@ public
   #   If you don't know whether a target string to parse is exactly 1 line or
   #   not, use CSV.parseLine instead of this method.
   #   
-  def CSV.parse( src )
+  def CSV.parse( src, colSep = ?, )
     nextIdx = 0
     resType = :DT_COLSEP
     cells = Row.new
     begin
       while ( resType.equal?( :DT_COLSEP ))
 	aCell = Cell.new
-	resType, nextIdx = parseBody( src, nextIdx, aCell )
+	resType, nextIdx = parseBody( src, nextIdx, aCell, colSep )
 	cells.push( aCell.isNull ? nil : aCell.data )
-	resType, nextIdx = parseSeparator( src, nextIdx )
+	resType, nextIdx = parseSeparator( src, nextIdx, colSep )
       end
     rescue IllegalFormatError
       return Row.new
@@ -615,11 +712,13 @@ public
   
 
   # SYNOPSIS
-  #   str = CSV.create( cells )
+  #   str = CSV.create( cells, colSep = ?, )
   #
   # ARGS
   #   cells: an Array of cell to be converted to CSV string.  Each cell must 
   #     respond to 'to_s'.
+  #   colSep: Column separator.  ?, by default.  If you want to separate
+  #     fields with semicolon, give ?; here.
   #
   # RETURNS
   #   str: a String of generated CSV string.
@@ -627,7 +726,7 @@ public
   # DESCRIPTION
   #   Create a line from cells.  Each cell is stringified by to_s.
   #   
-  def CSV.create( cells )
+  def CSV.create( cells, colSep = ?, )
     if ( cells.size == 0 )
       return ""
     end
@@ -640,18 +739,18 @@ public
 	else
 	  Cell.new( cells[ idx ].to_s, false )
 	end
-      createBody( cell, resStr )
+      createBody( cell, resStr, colSep )
       idx += 1
       if ( idx == cells.size )
 	break
       end
-      createSeparator( :DT_COLSEP, resStr )
+      createSeparator( :DT_COLSEP, resStr, colSep )
     end
     resStr
   end
   
   # SYNOPSIS
-  #   nofCell, idx = CSV.parseLine( src, idx, outDev )
+  #   nofCell, idx = CSV.parseLine( src, idx, outDev, colSep = ?, )
   #
   # ARGS
   #   src: a CSV data to be parsed.  Must respond '[]( idx )'.
@@ -660,6 +759,8 @@ public
   #     requirement.
   #   idx: index of parsing location of 'src'.  0 origin.
   #   outDev: buffer for parsed cells.  Must respond '<<( aCSV::Cell )'.
+  #   colSep: Column separator.  ?, by default.  If you want to separate
+  #     fields with semicolon, give ?; here.
   #
   # RETURNS
   #   nofCell: num of parsed cells.
@@ -679,15 +780,15 @@ public
   #	p parsed
   #   end while nofCells > 0
   #   
-  def CSV.parseLine( src, idx, outDev )
+  def CSV.parseLine( src, idx, outDev, colSep = ?, )
     nofCell = 0
     nextIdx = idx
     resType = :DT_COLSEP
     begin
       while ( resType.equal?( :DT_COLSEP ))
 	aCell = Cell.new
-	resType, nextIdx = parseBody( src, nextIdx, aCell )
-	resType, nextIdx = parseSeparator( src, nextIdx )
+	resType, nextIdx = parseBody( src, nextIdx, aCell, colSep )
+	resType, nextIdx = parseSeparator( src, nextIdx, colSep )
 	if resType.equal?( :DT_EOS )
 	  if nofCell > 0 || !aCell.isNull
 	    resType = :DT_ROWSEP
@@ -708,13 +809,15 @@ public
   end
   
   # SYNOPSIS
-  #   nofCells = CSV.createLine( src, cells, outDev )
+  #   nofCells = CSV.createLine( src, cells, outDev, colSep = ?, )
   #
   # ARGS
   #   src: an Array of CSV::Cell to be converted to CSV string.  Must respond to
   #     'size' and '[]( idx )'.  src[ idx ] must return CSV::Cell.
   #   cells: num of cells in a line.
   #   outDev: buffer for created CSV string.  Must respond to '<<( aString )'.
+  #   colSep: Column separator.  ?, by default.  If you want to separate
+  #     fields with semicolon, give ?; here.
   #
   # RETURNS
   #   nofCells: num of converted cells.
@@ -739,43 +842,46 @@ public
   #   end
   #   p buf
   #   
-  def CSV.createLine( src, cells, outDev )
+  def CSV.createLine( src, cells, outDev, colSep = ?, )
     srcSize = src.size
     if ( srcSize == 0 )
+      if cells == 0
+	createSeparator( :DT_ROWSEP, outDev, colSep )
+      end
       return 0
     end
     resType = :DT_COLSEP
     nofCells = 0
-    createBody( src[ nofCells ], outDev )
+    createBody( src[ nofCells ], outDev, colSep )
     nofCells += 1
     while (( nofCells < cells ) && ( nofCells != srcSize ))
-      createSeparator( :DT_COLSEP, outDev )
-      createBody( src[ nofCells ], outDev )
+      createSeparator( :DT_COLSEP, outDev, colSep )
+      createBody( src[ nofCells ], outDev, colSep )
       nofCells += 1
     end
     if ( nofCells == cells )
-      createSeparator( :DT_ROWSEP, outDev )
+      createSeparator( :DT_ROWSEP, outDev, colSep )
     else
-      createSeparator( :DT_COLSEP, outDev )
+      createSeparator( :DT_COLSEP, outDev, colSep )
     end
     nofCells
   end
   
   private
   class IllegalFormatError < RuntimeError; end
-  
+
   # Private class methods.
   class << self
   private
 
-    def parseBody( src, idx, aCell )
+    def parseBody( src, idx, aCell, colSep )
       nextIdx = idx
       quoted = false
       cr = false
       aCell.isNull = false
       state = :ST_START
       while (( c = src[ nextIdx ] ) && !state.equal?( :ST_END ))
-	if ( c == ?, )
+	if ( c == colSep )
 	  if state.equal?( :ST_START )
 	    aCell.isNull = true
 	    nextIdx -= 1
@@ -897,12 +1003,12 @@ public
       end
     end
   
-    def parseSeparator( src, idx )
+    def parseSeparator( src, idx, colSep )
       nextIdx = idx
       resType = :DT_FAIL
       if src[ nextIdx ].nil?
 	resType = :DT_EOS
-      elsif ( src[ nextIdx ] == ?, )
+      elsif ( src[ nextIdx ] == colSep )
 	nextIdx += 1
 	resType = :DT_COLSEP
       elsif (( src[ nextIdx ] == ?\r ) && ( src[ nextIdx + 1 ] == ?\n ))
@@ -915,10 +1021,12 @@ public
       return resType, nextIdx
     end
     
-    def createBody( cellData, outDev )
+    def createBody( cellData, outDev, colSep )
       addData = cellData.data.dup
       if ( !cellData.isNull )
-       	if ( addData.gsub!( '"', '""' ) || ( addData =~ /[,\"]/ ) ||
+       	if ( addData.gsub!( '"', '""' ) ||
+	    addData.include?( colSep ) ||
+	    addData.include?( ?" ) ||	# "
   	    ( addData =~ /\r\n/ ) || ( cellData.data.empty? ))
   	  outDev << '"' << addData << '"'
    	else
@@ -927,10 +1035,10 @@ public
       end
     end
     
-    def createSeparator( type, outDev )
+    def createSeparator( type, outDev, colSep )
       case type
       when :DT_COLSEP
-       	outDev << ','
+       	outDev << colSep.chr
       when :DT_ROWSEP
 	outDev << "\r\n"
       else
@@ -1133,21 +1241,6 @@ public
       return idxIsEOS?( 0 )
     end
   
-  protected
-    def terminate
-      while ( relBuf ); end
-    end
-  
-    # protected method 'read' must be defined in derived classes.
-    # CAUTION: Returning a string which size is not equal to 'size' means
-    #   EnfOfStream.  When it is not at EOS, you must block the callee, try to
-    #   read and return the sized string.
-    def read( size ) # raise EOFError
-      raise NotImplementedError.new( 'Method read must be defined in a derived class.' )
-    end
-  
-  private
-  
     # SYNOPSIS
     #   N/A
     #
@@ -1166,6 +1259,21 @@ public
       @curBuf = @bufTailIdx
     end
   
+  protected
+    def terminate
+      while ( relBuf ); end
+    end
+  
+    # protected method 'read' must be defined in derived classes.
+    # CAUTION: Returning a string which size is not equal to 'size' means
+    #   EnfOfStream.  When it is not at EOS, you must block the callee, try to
+    #   read and return the sized string.
+    def read( size ) # raise EOFError
+      raise NotImplementedError.new( 'Method read must be defined in a derived class.' )
+    end
+  
+  private
+  
     def addBuf
       if @isEOS
 	return false
@@ -1226,17 +1334,16 @@ public
   #
   class IOBuf < StreamBuf
   public
-    def close
-      terminate
-      @s.close
-    end
-
-  private
     def initialize( s )
       @s = s
       super()
     end
   
+    def close
+      terminate
+    end
+
+  private
     def read( size )
       @s.sysread( size )
     end
